from __future__ import annotations

import time
from datetime import datetime, timedelta
from typing import Optional

from stx.climate_controller import ClimateController
from stx.errors import ErrorResponse
from stx.memory_map import MemoryMap
from stx.plate_handler import PlateHandler
from stx.serial_port import SerialPort
from stx.shaker_controller import ShakerController


class Stx:
    __serial_port: SerialPort
    __memory_map: MemoryMap
    plate_handler: PlateHandler
    """
    Exposes functionality related to plate handling
    """
    shaker_controller: ShakerController
    """
    Exposes tower shaker functionality
    """
    climate_controller: ClimateController
    """
    Exposes functionality related to temperature and CO2
    """

    def __init__(self, serial_port: str):
        self.__serial_port = SerialPort(serial_port, timeout=1)
        self.__memory_map = MemoryMap(self.__serial_port)
        self.plate_handler = PlateHandler(self.__memory_map)
        self.shaker_controller = ShakerController(self.__memory_map)
        self.climate_controller = ClimateController(self.__memory_map)

    @property
    def ready(self) -> bool:
        """True if the device is not busy"""
        return self.__memory_map.bits[1915]

    @property
    def has_error(self) -> bool:
        """True if the device has an error"""
        return self.__memory_map.bits[1814]

    def initialize(self) -> None:
        """Initialize the device"""
        self.__memory_map.bits[1900] = True
        time.sleep(0.5)  # else it does not work
        self.__memory_map.bits[1801] = True

    def open_connection(self) -> None:
        """Open the connection"""
        self.__serial_port.open()

    def close_connection(self) -> None:
        """Close the connection"""
        self.__serial_port.close()

    def get_error(self) -> Optional[ErrorResponse]:
        """If the device has an error, return it. Else, return None."""
        if self.has_error:
            return ErrorResponse.from_error_code(self.__memory_map.data[200])
        return None

    def wait_until_ready(self, timeout: float) -> None:
        """
        Block the current thread until the device is not busy anymore

        Parameters
        ----------
        timeout
            Timeout in seconds

        Raises
        ------
        TimeoutError
            If the device is still busy after the timeout
        ErrorResponse
            If the device has an error
        """
        end_time = datetime.now() + timedelta(seconds=timeout)
        while not (self.ready or self.has_error):
            if datetime.now() > end_time:
                raise TimeoutError(f"Device still busy after {timeout} seconds")

        error = self.get_error()
        if error is not None:
            raise error

    def __enter__(self) -> Stx:
        self.open_connection()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close_connection()



from __future__ import annotations


class SerialError(Exception):
    """Error during serial communication. Corresponds to the responses E0 to E5."""

    @staticmethod
    def from_response(response: str) -> SerialError:
        error_map = {
            "E0": RelayError,
            "E1": CommandError,
            "E2": ProgramError,
            "E3": HardwareError,
            "E4": WriteProtectedError,
            "E5": BaseUnitError,
        }
        if response in error_map:
            return error_map[response]()
        return SerialError(f"Unexpected response: '{response}'")


class RelayError(SerialError):
    def __init__(self):
        super().__init__("Undefined timer, counter, data memory. Check if requested unit is valid.")


class CommandError(SerialError):
    def __init__(self):
        super().__init__(
            "Invalid command. Check if communication is opened by CR, check command sent to controller, "
            "check for interruptions during string transmission."
        )


class ProgramError(SerialError):
    def __init__(self):
        super().__init__("Firmware lost, reprogram controller")


class HardwareError(SerialError):
    def __init__(self):
        super().__init__(
            "Controller hardware error, turn controller ON/OFF, controller is faulty and has to be replaced."
        )


class WriteProtectedError(SerialError):
    def __init__(self):
        super().__init__("Unauthorized access")


class BaseUnitError(SerialError):
    def __init__(self):
        super().__init__("Unauthorized access")


class ErrorResponse(Exception):
    """Error during command execution. Corresponds to the data at address DM200."""

    @staticmethod
    def from_error_code(error_code: int) -> ErrorResponse:
        error_code = int(hex(error_code)[-2:], base=16)  # convert to hex, take two last digits, parse as hex
        messages = {
            0x01: "General time out; Operation took longer than 15 Minutes",
            0x02: "General time out; Time out after previous time out; Operation longer than 5 Minutes",
            0x03: "Init time out; Initialization has taken too long",
            0x04: "Carousel miss count; Number of cassette positions during init wrong",
            0x07: "Gate OPEN time out",
            0x08: "Gate CLOSE time out",
            0x09: "Lift travel exceeds maximum value",
            0x0A: "Wrong Cassette; DM0 > DM29; Cassette number exceeds maximum cassette",
            0x0B: "Lift overflow; travel path exceeds maximum allowed path",
            0x0C: "Wrong Level; DM5 > DM25; Level does exceed the maximum available levels",
            0x0D: "Plate trace Error; Plate was not loaded/unloaded as expected from/to shovel",
            0x0E: "Init time out; System was not able to initilize",
            0x10: "Turn in Turn Init Sensor or not in safe position; Possible step loss",
            0x12: "Carousel Init time out",
            0x13: "Shovel OUT time out; shovel could not be extended",
            0x14: "Shovel IN time out; Shovel could not be retracted",
        }
        if error_code not in messages:
            return ErrorResponse(f"Unknown error code: {error_code}")
        return ErrorResponse(messages[error_code])



from typing import Optional

from stx.memory_map import MemoryMap


class PlateHandler:
    __memory_map: MemoryMap

    def __init__(self, memory_map: MemoryMap) -> None:
        self.__memory_map = memory_map

    @property
    def transfer_station_occupied(self) -> bool:
        """True if the transfer station is occupied, False otherwise"""
        return self.__memory_map.bits[1813]

    @property
    def handler_occupied(self) -> bool:
        """True if the plate handler shovel is occupied, False otherwise"""
        return self.__memory_map.bits[1812]

    def rotate_to_stacker(self, stacker: int) -> None:
        """
        Rotate the carrousel to the given stacker

        Parameters
        ----------
        stacker
            The stacker number
        """
        self.__memory_map.data[0] = stacker

    def set_current_slot(self, slot: int) -> None:
        """
        Set the current slot

        Parameters
        ----------
        slot
            The slot number
        """
        self.__memory_map.data[5] = slot

    def move_plate_from_transfer_station_to_slot(self, stacker: int, slot: int) -> None:
        """
        Move a plate from the transfer station to the given slot

        Parameters
        ----------
        stacker
            The stacker number
        slot
            The slot number
        """
        self.__move_plate(stacker, slot, 1904)

    def move_plate_from_slot_to_transfer_station(self, stacker: int, slot: int) -> None:
        """
        Move a plate from the given slot to the transfer station

        Parameters
        ----------
        stacker
            The stacker number
        slot
            The slot number
        """
        self.__move_plate(stacker, slot, 1905)

    def move_plate_from_handler_to_transfer_station(self) -> None:
        """Move a plate from the plate handler shovel to the transfer station"""
        self.__move_plate(None, None, 1906)

    def move_plate_from_transfer_station_to_handler(self) -> None:
        """Move a plate from the transfer station to the plate handler shovel"""
        self.__move_plate(None, None, 1907)

    def move_plate_from_slot_to_handler(self, stacker: int, slot: int) -> None:
        """
        Move a plate from the given slot to the plate handler shovel

        Parameters
        ----------
        stacker
            The stacker number
        slot
            The slot number
        """
        self.__move_plate(stacker, slot, 1908)

    def move_plate_from_handler_to_slot(self, stacker: int, slot: int) -> None:
        """
        Move a plate from the plate handler shovel to the given slot

        Parameters
        ----------
        stacker
            The stacker number
        slot
            The slot number
        """
        self.__move_plate(stacker, slot, 1909)

    def __move_plate(self, stacker: Optional[int], slot: Optional[int], action_bit: int):
        """Helper method for plate movement"""
        # if not specified, use current stacker and slot. Undefined (0) after initialization -> set to 1
        if stacker is None:
            stacker = self.__memory_map.data[0]
            if stacker == 0:
                stacker = 1
        if slot is None:
            slot = self.__memory_map.data[5]
            if slot == 0:
                slot = 1

        self.rotate_to_stacker(stacker)
        self.set_current_slot(slot)
        self.__memory_map.bits[action_bit] = True



from __future__ import annotations

from stx.serial_port import SerialPort


class MemoryMap:
    """
    Interface to the bit and data registers

    Examples
    --------
    >>> memory_map = MemoryMap(SerialPort("COM3"))

    >>> memory_map.bits[1915]  # read bit 1915 (ready)
    True
    >>> memory_map.bits[1801] = True  # set bit 1801 (initialize)

    >>> memory_map.data[982]  # read data DM982 (actual temperature in 0.1 °C)
    373
    >>> memory_map.data[890] = 400  # set data DM890 (target temperature in 0.1 °C)
    """

    bits: BitMemoryMap
    data: DataMemoryMap

    def __init__(self, serial_port: SerialPort):
        self.bits = BitMemoryMap(serial_port)
        self.data = DataMemoryMap(serial_port)


class DataMemoryMap:
    """Interface to the data memory"""

    __serial_port: SerialPort

    def __init__(self, serial_port: SerialPort):
        self.__serial_port = serial_port

    def __getitem__(self, address: int) -> int:
        return self.__serial_port.read_data(address)

    def __setitem__(self, address: int, value: int):
        self.__serial_port.send_and_expect_ok(f"WR DM{address} {value}")


class BitMemoryMap:
    """Interface to the bits"""

    __serial_port: SerialPort

    def __init__(self, serial_port: SerialPort):
        self.__serial_port = serial_port

    def __getitem__(self, address: int) -> bool:
        return self.__serial_port.read_bit(address)

    def __setitem__(self, address: int, value: bool):
        operator = "ST" if value else "RS"
        self.__serial_port.send_and_expect_ok(f"{operator} {address}")



from __future__ import annotations

import time
from threading import Lock

from serial import PARITY_EVEN, Serial

from stx.errors import SerialError
from stx.utils import lock_threading_lock


class SerialPort:
    __serial_port: Serial
    __port_lock: Lock
    __timeout: float

    def __init__(self, port: str, *, timeout: float) -> None:
        """
        Wraps a plain serial connection to simplify communication with LiCONiC STX devices

        Parameters
        ----------
        port : str
            The address of the port (e.g. "COM3" on Windows or "/dev/ttyACM0" on Unix)
        timeout: float, seconds
            A :class:`TimeoutError` will be raised of read or write operations took longer than this duration
        """
        self.__serial_port = Serial(port, parity=PARITY_EVEN, timeout=timeout, write_timeout=timeout)
        self.__port_lock = Lock()
        self.__timeout = timeout

    def __communicate(self, command: str) -> str:
        """
        Send a command to the port and return the response

        Raises
        ------
        TimeoutError
            If a read or write operation took too long
        SerialError
            If the response was E0 to E5, or there were unread bytes in the input buffer before sending the command
        """
        with lock_threading_lock(self.__port_lock, timeout=self.__timeout):
            if self.__serial_port.in_waiting:
                raise SerialError("There were unread bytes in the input buffer")

            raw_response = b""
            self.__serial_port.write(command.encode("ascii") + b"\r")

            time.sleep(0.1)
            while not raw_response.endswith(b"\r\n"):
                char = self.__serial_port.read()
                if not char:
                    raise TimeoutError(rf"Did not receive a '\r\n'-terminated response after {self.__timeout} seconds")
                raw_response += char

            response = raw_response[:-2].decode("ascii")

            if response in ("E0", "E1", "E2", "E3", "E4", "E5"):
                raise SerialError.from_response(response)

            return response

    def open(self) -> None:
        """Open the connection by sending 'CR'"""
        self.send_and_expect_response("CR", "CC")

    def close(self) -> None:
        """Close the connection by sending 'CQ'"""
        self.send_and_expect_response("CQ", "CF")

    def send_and_expect_response(self, command: str, expected_response: str) -> None:
        """
        Send the command and raise a SerialError if the response was different than expected

        Parameters
        ----------
        command
            The command to send to the device
        expected_response
            The expected response

        Raises
        ------
        SerialError
            If the response was different than expected
        """
        response = self.__communicate(command)
        if response != expected_response:
            raise SerialError.from_response(response)

    def send_and_expect_ok(self, command: str) -> None:
        """
        Send the command and raise a SerialError if the response was not 'OK'

        Parameters
        ----------
        command
            The command to send to the device

        Raises
        ------
        SerialError
            If the response was different than expected
        """
        self.send_and_expect_response(command, "OK")

    def read_bit(self, address: int) -> bool:
        """
        Read the bit value at the given address

        Parameters
        ----------
        address
            The address to read
        """
        return self.__communicate(f"RD {address}") == "1"

    def read_data(self, address: int) -> int:
        """
        Read the data memory value at the given address

        Parameters
        ----------
        address
            The address to read
        """
        return int(self.__communicate(f"RD DM{address}"))



from stx.memory_map import MemoryMap


class ClimateController:
    __memory_map: MemoryMap

    def __init__(self, memory_map: MemoryMap) -> None:
        self.__memory_map = memory_map

    @property
    def current_temperature(self) -> float:
        """The current temperature in °C"""
        return self.__memory_map.data[982] / 10

    @property
    def target_temperature(self) -> float:
        """The target temperature in °C (can be set)"""
        return self.__memory_map.data[890] / 10

    @target_temperature.setter
    def target_temperature(self, target_temperature: float) -> None:
        self.__memory_map.data[890] = int(target_temperature * 10)

    @property
    def current_humidity(self) -> float:
        """The current humidity in % RH"""
        return self.__memory_map.data[983] / 10

    @property
    def target_humidity(self) -> float:
        """The target humidity in % RH (can be set)"""
        return self.__memory_map.data[893] / 10

    @target_humidity.setter
    def target_humidity(self, target_humidity: float) -> None:
        self.__memory_map.data[893] = int(target_humidity * 10)

    @property
    def current_co2(self) -> float:
        """The current CO2 level in % vol."""
        return self.__memory_map.data[984] / 100

    @property
    def target_co2(self) -> float:
        """The current CO2 level in % vol. (can be set)"""
        return self.__memory_map.data[894] / 100

    @target_co2.setter
    def target_co2(self, target_co2: float) -> None:
        self.__memory_map.data[894] = int(target_co2 * 100)

    @property
    def current_gas_1(self) -> float:
        """The current gas level of gas port 1 in % vol. (N2 or O2)"""
        return self.__memory_map.data[985] / 100

    @property
    def target_gas_1(self) -> float:
        """The target gas level of gas port 1 in % vol. (N2 or O2, can be set)"""
        return self.__memory_map.data[895] / 100

    @target_gas_1.setter
    def target_gas_1(self, target_n2: float) -> None:
        self.__memory_map.data[895] = int(target_n2 * 100)

    @property
    def current_gas_2(self) -> float:
        """The current gas level of gas port 2 in % vol. (N2 or O2)"""
        return self.__memory_map.data[986] / 100

    @property
    def target_gas_2(self) -> float:
        """The target gas level of gas port 2 in % vol. (N2 or O2, can be set)"""
        return self.__memory_map.data[896] / 100

    @target_gas_2.setter
    def target_gas_2(self, target_n2: float) -> None:
        self.__memory_map.data[896] = int(target_n2 * 100)

    def reset_water_low_alarm(self) -> None:
        """Reset the "water low" alarm"""
        self.__memory_map.bits[1505] = False

    def reset_co2_timeout_error(self) -> None:
        """Reset the CO2 timeout error"""
        self.__memory_map.bits[1504] = False



from stx.memory_map import MemoryMap


class ShakerController:
    __memory_map: MemoryMap

    def __init__(self, memory_map: MemoryMap) -> None:
        self.__memory_map = memory_map

    @property
    def shaker_is_active(self) -> bool:
        """True if the shaker is active, False otherwise"""
        return self.__memory_map.bits[1913]

    def activate_shaker(self) -> None:
        """Activate the shaker"""
        self.__memory_map.bits[1913] = True

    def stop_shaker(self) -> None:
        """Stop the shaker"""
        self.__memory_map.bits[1913] = False

    @property
    def shaker_speed(self) -> int:
        """The shaker speed, between 1 and 50 (can be set)"""
        return self.__memory_map.data[39]

    @shaker_speed.setter
    def shaker_speed(self, speed: int) -> None:
        self.__memory_map.data[39] = speed



from contextlib import contextmanager
from enum import IntEnum
from threading import Lock
from typing import Iterator, TypeVar

GenericIntEnum = TypeVar("GenericIntEnum", bound=IntEnum)


@contextmanager
def lock_threading_lock(lock: Lock, *, timeout: float) -> Iterator[None]:
    """Context manager for using a `threading.lock` with a timeout"""
    if not timeout > 0:
        raise ValueError("Timeout must be positive")

    if not lock.acquire(timeout=timeout):
        raise TimeoutError(f"Could not acquire lock after {timeout} seconds")

    try:
        yield
    finally:
        lock.release()
